module problem_mod
	use omp_lib
	implicit none

	! Boundary value at the LHS of the pipe
	real(kind=8), parameter :: LEFT_VALUE = 1.0
	! Boundary value at the RHS of the pipe
	real(kind=8), parameter :: RIGHT_VALUE = 10.0
	! The maximum number of iterations
	integer, parameter :: MAX_ITERATIONS = 100000
	! How often to report the norm
	integer, parameter :: REPORT_NORM_PERIOD = 100

contains

	subroutine run_solver()
		real(kind=8), dimension(:,:), allocatable :: u_k, u_kp1, temp
		real(kind=8) :: bnorm, rnorm, norm, convergence_accuracy
		integer :: i, j, k, ierr, nx, ny, max_its, start_time_values(8), end_time_values(8)
		integer(kind=8) :: start_ms, end_ms
		character(len=32) :: arg
		double precision :: start_time

		if (command_argument_count() /= 4) then
			print *, &
				"You must provide four command line arguments, the global size in X,", &
				"the global size in Y, convergence accuracy and max number iterations"
			return
		end if

		call get_command_argument(1, arg)
		read(arg,*) nx
		call get_command_argument(2, arg)
		read(arg,*) ny
		call get_command_argument(3, arg)
		read(arg,*) convergence_accuracy
		call get_command_argument(4, arg)
		read(arg,*) max_its

		print *, "Global size in X=", nx, "Global size in Y=", ny

		allocate(u_k(0:ny+1, 0:nx+1), u_kp1(0:ny+1, 0:nx+1))

		bnorm=0.0
		rnorm=0.0
    
		call initialise_values(u_k, u_kp1, nx, ny)
		
		! Calculate the initial residual
		do i=1, nx
			do j=1, ny
				bnorm=bnorm+((u_k(j,i)*4-u_k(j-1,i)-u_k(j+1,i)-u_k(j,i-1)-u_k(j,i+1)) ** 2)
			end do
		end do		
		bnorm=sqrt(bnorm)

		call date_and_time(values=start_time_values)
		do k=0, MAX_ITERATIONS      			
			! Calculates the current residual norm
			do i=1, nx
				do j=1, ny
					rnorm=rnorm+((u_k(j,i)*4-u_k(j-1,i)-u_k(j+1,i)-u_k(j,i-1)-u_k(j,i+1)) ** 2)
				end do
			end do
			! Do the Jacobi iteration
			do i=1, nx
				do j=1, ny
					u_kp1(j,i)=0.25 * (u_k(j-1,i) + u_k(j+1,i) + u_k(j,i-1) + u_k(j,i+1))
				end do
			end do

      			call move_alloc(u_kp1, temp)
            		call move_alloc(u_k, u_kp1)
            		call move_alloc(temp, u_k)

      			norm=sqrt(rnorm)/bnorm
      			rnorm=0.0

			if (norm .lt. convergence_accuracy) exit
			if (max_its .gt. 0 .and. k .ge. max_its) exit


      			if (mod(k, REPORT_NORM_PERIOD)==0) print *, "Iteration=",k," Relative Norm=",norm
		end do
		call date_and_time(values=end_time_values)
		print *, "Terminated on ",k," iterations, Relative Norm=", norm
		print *, "Total runtime is ", getMSFromTime(end_time_values) - getMSFromTime(start_time_values), " ms"
		deallocate(u_k, u_kp1)
	end subroutine run_solver

	! Initialises the arrays, such that u_k contains the boundary conditions at the start and end points and all other
	! points are zero. u_kp1 is set to equal u_k
	subroutine initialise_values(u_k, u_kp1, nx, ny)
		real(kind=8), intent(inout) :: u_k(0:ny+1, 0:nx+1), u_kp1(0:ny+1, 0:nx+1)
		integer, intent(in) :: nx, ny

		! We are setting the boundary (left and right) values here, in the parallel version this should be exactly the same and no changed required

		u_k(0,:)=LEFT_VALUE
		u_k(ny+1,:)=RIGHT_VALUE
		u_k(1:ny,:)=0.0_8
		u_kp1=u_k
	end subroutine initialise_values  

	integer(kind=8) function getMSFromTime(time_values)
		integer :: time_values(8)

		getMSFromTime = ( time_values(5) )*60
	 	getMSFromTime = ( getMSFromTime + time_values(6) )*60
		getMSFromTime = ( getMSFromTime + time_values(7) )*1e3
		getMSFromTime = getMSFromTime + time_values(8) 
	end function getMSFromTime
end module problem_mod

program diffusion
	use problem_mod
	implicit none

	call run_solver()
end program diffusion
